#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/fs.h>
#include <linux/mutex.h>
#include <linux/slab.h>
#include <linux/uaccess.h>

static struct usb_driver androidbulk_driver;
/**
 * ドライバがUSBデバイスを扱うための構造体
 *
 * @udev USBコアで管理するUSBデバイス情報
 * @bulkout_buffer データ転送用バッファ
 * @bulk_out_endpointAddr バルク転送用エンドポイントアドレス
 * @max_buffer_size バルク転送用エンドポイントの最大パケットサイズ
 */
struct androidbulk_dev {
	struct usb_device *udev;
	char*	bulkout_buffer;
	__u8	bulk_out_endpointAddr;
	unsigned int max_buffer_size;
};

/**
 * データ転送処理と USB デバイス切断処理との競合を
 * 回避するためのロック
 */
static DEFINE_MUTEX(androidbulk_mutex);

/**
 * read は未サポート
 */
static ssize_t
androidbulk_read(struct file *file, char __user *buf, size_t len, loff_t *loff)
{
	return -EPERM;
}

/**
 * USBデバイスへのデータ転送関数
 *
 * @file file 情報
 * @buf データ転送用のバッファ(ユーザ空間のメモリ領域)
 * @len データ転送サイズ
 * @loff ファイルオフセット
 * @return 成功 実際に転送したサイズ、失敗 負の値
 */
static ssize_t
androidbulk_write(struct file *file, const char __user *buf, size_t len, loff_t *loff)
{
	size_t to_write;
	struct androidbulk_dev *dev = file->private_data;
	int ret;
	int actual_length;
	size_t total_writed;

	total_writed = 0;
	while (len > 0) {
		to_write = min(len, dev->max_buffer_size);

		if (copy_from_user(dev->bulkout_buffer, buf+total_writed, to_write)) {
			printk(KERN_ERR "androidbulk:copy_from_user failed!\n");
			return -EFAULT;
		}

		ret = usb_bulk_msg(dev->udev, usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr),
				dev->bulkout_buffer, to_write, &actual_length, 3*HZ);
		if (ret || actual_length != to_write) {
			printk(KERN_ERR "androidbulk:usb_bulk_msg failed!\n");
			return -EFAULT;
		}
		len -= to_write;
		total_writed += to_write;
	}
	return total_writed;
}

/**
 * USB デバイスのオープン関数
 *
 * データ転送処理の開始契機であり、androidbulk_mutex のロックを獲得する。
 * ロック獲得できない場合は、以下が動作している場合であり、エラー終了する。
 *
 * ・他データ転送処理が実行中
 * ・USB切断処理中
 *
 * <注意点>
 * データ転送中に獲得するロック範囲は、設計をシンプルにするために広くしている。
 * より汎用的な設計を目指す場合は、ロック範囲を局所化しなければならない。
 *
 * @node USBデバイスのinode
 * @file USBデバイスのfile
 * @return 成功 0、失敗 0以外
 */
static int
androidbulk_open(struct inode *node, struct file *file)
{
	struct usb_interface *interface;
	struct androidbulk_dev *dev;

	// ロック獲得できない場合は、エラー終了する。
	if (!mutex_trylock(&androidbulk_mutex)) {
		return -EBUSY;
	}

	// インタフェース情報を取得する。
	interface = usb_find_interface(&androidbulk_driver, iminor(node));
	if (!interface) {
		mutex_unlock(&androidbulk_mutex);
		return -ENODEV;
	}

	// USBデバイス情報を取得する。
	dev = usb_get_intfdata(interface);
	if (!dev) {
		mutex_unlock(&androidbulk_mutex);
		return -ENODEV;
	}

	// file のプライベート情報としてUSBデバイス情報を設定しておく。
	file->private_data = dev;
	return 0;
}

/**
 * USB デバイスのクローズ関数
 *
 * データ転送処理の終了契機であり、androidbulk_mutex のロックを解除する。
 *
 * @node USBデバイスのinode
 * @file USBデバイスのfile
 * @return 0(常に成功する )
 */
static int
androidbulk_release(struct inode *node, struct file *file)
{
	mutex_unlock(&androidbulk_mutex);
	return 0;
}

/**
 * USBデバイスにデータ転送するためのファイルI/O関数群
 *
 * @owner 本モジュール
 * @read 未サポート
 * @write USBデバイスへのデータ転送関数
 * @open USBデバイスのオープン関数
 * @release USBデバイスのクローズ関数
 */
static struct file_operations androidbulk_fops = {
	.owner 	=	THIS_MODULE,
	.read 	=	androidbulk_read,
	.write	=	androidbulk_write,
	.open   =	androidbulk_open,
	.release=	androidbulk_release,
};

/**
 * ドライバが USB コアに登録するUSB デバイス情報
 *
 * @name デバイスファイル名(名前にデバイス番号を含めているので%dを付加している)
 * @fops データ転送用のファイルI/O関数群
 * @minor_base このドライバに割り当てる先頭のマイナー番号
 *
 */
static struct usb_class_driver androidbulk_class = {
	.name = 	"androidbulk%d",
	.fops =		&androidbulk_fops,
	.minor_base=	100,
};

/**
 * USBコアが、(本ドライバで扱う)USBデバイスの切断を検出した際に呼び出す関数
 * 本関数では、USBデバイスの登録を解除すると共に、デバイスファイルを削除する。
 *
 * <競合時の考慮点>
 * データ転送中にUSBデバイスの切断の検出が発生すると、競合が発生する。
 * このため、データ転送処理と disconnect 処理は androidbulk_mutex のロック
 * でシリアライズする。
 *
 * @param interface USB デバイスのインタフェース情報
 */
static void
androidbulk_disconnect(struct usb_interface *interface)
{
	struct androidbulk_dev *dev = NULL;

	mutex_lock(&androidbulk_mutex);
	printk(KERN_INFO "androidbulk:androidbulk disconnected!\n");
	dev = usb_get_intfdata(interface);
	if(dev) {
		kfree(dev);
		dev = NULL;
	}
	usb_set_intfdata(interface, NULL);
	usb_deregister_dev(interface, &androidbulk_class);
	mutex_unlock(&androidbulk_mutex);
	return;
}

/**
 * USBコアが、(本ドライバで扱う)USBデバイスの接続を検出した際に呼び出す関数
 * 本関数では、USB デバイスのエンドポイント情報の検査をし、
 * デバイスファイルを作成する。
 *
 * @param interface USB デバイスのインタフェース情報
 * @param id ドライバが扱う USB デバイス情報
 * @return 成功 0、失敗 0 以外
 */
static int
androidbulk_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	int ret;
	struct androidbulk_dev *dev = NULL;
	struct usb_host_interface *iface_desc = NULL;
	struct usb_endpoint_descriptor *endpoint = NULL;
	int i;

	printk(KERN_INFO "androidbulk:androidbulk probing...\n");

	/*
	 * USB デバイスのエンドポイントを検査する。
	 * 本ドライバではバルク転送方式で、転送方向が OUT のものを検索する。
	 * また、最大パケットサイズをデータ転送用のバッファサイズとして、
	 * 転送用バッファのメモリを確保する。
	 */
	iface_desc = interface->cur_altsetting;
	for(i=0; i < iface_desc->desc.bNumEndpoints; i++) {
		endpoint = &(iface_desc->endpoint[i].desc);
		if (usb_endpoint_is_bulk_out(endpoint)) {
			printk(KERN_INFO "androidbulk:bulk out endpoint found!\n");
			dev = kzalloc(sizeof(*dev), GFP_KERNEL);
			if (!dev) {
				ret = -ENOMEM;
				goto errdone;
			}
			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
			dev->max_buffer_size = endpoint->wMaxPacketSize;
			dev->bulkout_buffer = kzalloc(dev->max_buffer_size, GFP_KERNEL);
			if (!(dev->bulkout_buffer)) {
				ret = -ENOMEM;
				goto errdone;
			}
			break;
		}
	}

	/*
	 * 本ドライバが扱うエンドポイントが存在しない場合は、エラー終了する。
	 */
	if (!dev) {
		ret = -EBUSY;
		goto errdone;
	}

	/*
	 * USBデバイスを登録すると共に、デバイスファイルを作成する。
	 */
	ret = usb_register_dev(interface, &androidbulk_class);
	if(ret) {
		printk(KERN_ERR "androidbulk: usb_register_dev failed!\n");
		return ret;
	}
	/*
	 * デバイス情報にUSBコアのデバイス情報を設定しておく。
	 * ここで設定した情報は、ファイルI/Oする際に使用する。
	 */
	dev->udev = usb_get_dev(interface_to_usbdev(interface));

	/*
	 * インタフェース情報にUSBデバイス情報を設定しておく。
	 * ここで設定した情報は、ファイルI/Oする際に使用する。
	 */
	usb_set_intfdata(interface, dev);

	return 0;

errdone:
	if(dev) {
		if (dev->bulkout_buffer) {
			kfree(dev->bulkout_buffer);
			dev->bulkout_buffer = NULL;
		}
		kfree(dev);
		dev = NULL;
	}
	return ret;
}

/**
 * 本ドライバが扱うUSBデバイス情報
 *
 * ベンダID 0x4E8
 * プロダクトID 0x1234
 */
static struct usb_device_id androidbulk_table[]= {
	{ USB_DEVICE(0x4E8, 0x1234)},
	{ }
};

/**
 * USB デバイスドライバを USB コアに登録する情報
 *
 * @name ドライバの名前
 * @probe USBコアが、(本ドライバで扱う)USBデバイスの接続を検出した際に呼び出す関数
 * @disconnect USBコアが、(本ドライバで扱う)USBデバイスの切断を検出した際に呼び出す関数
 * @id_table 本ドライバが扱うUSBデバイス情報
 * @supports_autosuspend 本ドライバではサスペンドは未対応
 */
static struct usb_driver androidbulk_driver= {
	.name=		"androidbulk",
	.probe=		androidbulk_probe,
	.disconnect=	androidbulk_disconnect,
	.id_table=	androidbulk_table,
	.supports_autosuspend=0,
};

/**
 * USB デバイスドライバを Linux に登録する関数
 *
 * 本関数は、ユーザが insmod コマンドで本モジュールをインストール
 * する際に呼び出される。
 */
static int __init
androidbulk_init(void)
{
	int result;
	printk(KERN_INFO "androidbulk:androidbulk loaded\n");
	result = usb_register(&androidbulk_driver);
	if(result)	{
		printk(KERN_ERR "androidbulk:usb_register failed: %d", result);
		return result;
	}
	return 0;
}

/**
 * USB デバイスドライバを Linux から登録解除する関数
 *
 * 本関数は、ユーザが rmmod コマンドで本モジュールをアンインストール
 * する際に呼び出される。
 *
 * <競合時の考慮点>
 * データ転送中にユーザがアンインストールすると、競合が発生する。
 * usb_deregister() では、USB コアで USB デバイスとの接続を切り、
 * 本ドライバの disconnect() が呼び出される。
 * このため、disconnect() 関数にて、ユーザとデータ転送処理との間で
 * 競合が発生した場合の考慮を行うことにする。
 *
 */
static void __exit
androidbulk_exit(void)
{
	usb_deregister(&androidbulk_driver);
	printk(KERN_INFO "androidbulk:androidbulk unloaded\n");
}

module_init(androidbulk_init);
module_exit(androidbulk_exit);
MODULE_LICENSE("GPL");
